// TO COMPILE:w
// gcc -DDOUBLE_PRECISION -fPIC -shared *.cpp -o lib_CPP.so

#include "uclib.h"
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <vector>
#include <numeric>  // for std::accumulate  

#define ENABLE_DATA_TRANSFER // possible data transfer vector for trouble shooting - can be disabled
// #define VERBOSE // enables more print output

typedef enum {  
    CH4, CO, CO2, H2, H2O, N2, O2, OH, 							// fluid phase species, 
	CS, CHS, CH2SS, CH3S, COS, CO2S, HS, H2OS, OS, OHS, PtS, 	// surface site species, 
	PtB															// bulk species IN CORRECT ORDER
} SpeciesOrder;  
typedef enum {  
    _H22PTSto2HS, _2HStoH22PTS, _O22PTSto2OS, _2OStoO22PTS, _H2OPTStoH2OS, _H2OStoH2OPTS, _OHPTStoOHS, _OHStoOHPTS, 
	_HSOStoOHSPTS, _HSOHStoH2OSPTS, _OHSOHStoH2OSOS, _COPTStoCOS, _COStoCOPTS, _CO2StoCO2PTS, _COSOStoCO2SPTS, 
	_CH42PTStoCH3SHS, _CH3SPTStoCH2SSHS, _CH2SSPTStoCHSHS, _CHSPTStoCSHS, _CSOStoCOSPTS, _COSPTStoCSOS
} ReactionOrder;  
typedef enum {  
	_CellIndex, _onOff
} FunctionOrder;  


#ifdef ENABLE_DATA_TRANSFER // store cell ID and the data in vectores for transfer to a user field function
#include <unordered_map> 
std::unordered_map<double, double> transferMap;  
#endif

// zi: A pointer to a vector (type double, size nSpe) of gas phase specific mole fractions, surface site fractions, bulk activities, and cell temperature.
// e.g. zi[nSpe] is the cell temperature, T provides the wall temperature!
void reactionRateModify(double *reacRateForward, double *reacRateBackward, double T, unsigned int nSpe, double *zi, void *data)
{
	struct UserAccessibleData *udata = (struct UserAccessibleData *)data;
	unsigned int nReactions = udata->_reactionData._nReactions;
	unsigned int nFFs = udata->_fieldFunctionsData._nFF;
	double p =  udata->_pressure; 								// absolute pressure
	double *f =  udata->_fieldFunctionsData._ffVal;
	unsigned int nGas = udata->_fluidPhaseProperties._nGas;
	unsigned int nSite = udata->_surfaceChemistryData._nSite;
	unsigned int nBulk = udata->_surfaceChemistryData._nBulk;

	// double *M =  udata->_fluidPhaseProperties._molecularWeight;
	// Copy the molecular weights from the individual arrays to the concatenated array matching the species order enum
	double* M = new double[nGas + nSite + nBulk];  
	std::copy(udata->_fluidPhaseProperties._molecularWeight, udata->_fluidPhaseProperties._molecularWeight + nGas, M);  
	std::copy(udata->_surfaceMaterialProperties._siteMolecularWeight, udata->_surfaceMaterialProperties._siteMolecularWeight + nSite, M + nGas);  
	std::copy(udata->_surfaceMaterialProperties._bulkMolecularWeight, udata->_surfaceMaterialProperties._bulkMolecularWeight + nBulk, M + nGas + nSite);  
	
	double* wi = new double[nGas];   						//gas mass fractions
    for (unsigned int i = 0; i < nGas; ++i) 
        wi[i] = zi[i] * M[i];  
	double moleSum = std::accumulate(zi, zi + nGas, 0.0);  
	double* xi = new double[nGas];   						//gas mole fractions
    for (unsigned int i = 0; i < nGas; ++i) 
        xi[i] = zi[i] / moleSum;  
	double* pi = new double[nGas];   						//gas partial pressures
    for (unsigned int i = 0; i < nGas; ++i)   
        pi[i] = p * zi[i] / moleSum;  
	
	// testing user input function. this simply turns reactions off using the _onOff user function
	for (unsigned int i = 0; i < nReactions; ++i)	{
		reacRateForward[i] 	*= f[_onOff];
		reacRateBackward[i] *= f[_onOff]; 
	}

	// adjusting reaction rate for one reaction: wCH4+2PT(S)=>CH3(S)+H(S)
	reacRateForward[_CH42PTStoCH3SHS] 	*= 1/(1+pi[CH4]);
	// reacRateBackward[_CH42PTStoCH3SHS] 	= .01*1/(1+p*zi[CH4]*M[CH4]); // there is no back reaction here

	#ifdef ENABLE_DATA_TRANSFER 	// only enable for data transfer for debugging - slows down code 
        transferMap[f[_CellIndex]] = 1 / (1 + pi[CH4]);  
	#endif

	#ifdef VERBOSE
	// limit the output a little by only printing for cellid 0. this still prints a lot for each CVODE iterations
	if (f[_CellIndex] == 0.0 )
    	printf ("P %f, T %f, pCO %f, rCH4 %f, bCH4 %f, iCH4 %u, nReac %u, nSpe %u, nFFs %u \n", 

	// RUN THIS ONCE TO GET THE SPECIES ORDER - the az09 method is defined in uclib.h
	printf("SPECIES ORDER:\n");  
	for (unsigned int i = 0; i < nGas; i++) {  
		printf("%s, ", az09(udata->_fluidPhaseProperties._speciesName[i]));  
	}  
	printf("\n");     	
	for (unsigned int i = 0; i < nSite; i++) {  
		printf("%s, ", az09(udata->_surfaceMaterialProperties._siteSpeciesName[i]));  
	}  
	printf("\n");  
	for (unsigned int i = 0; i < nBulk; i++) {  
		printf("%s, ", az09(udata->_surfaceMaterialProperties._bulkSpeciesName[i]));  
	}  
	printf("\nFIELD FUNCTIONS ORDER:\n");  
	for (unsigned int i = 0; i < nFFs; i++) {
		printf("%s, ", az09(udata->_fieldFunctionsData._ffName[i]));
	}
	printf("\nREACTIONS ORDER:\n");  
	for (unsigned int i = 0; i < nReactions; i++) {
		printf("%s, ", az09(udata->_reactionPropertyDataGeneral._reactionName[i]));
	}
	printf("\n");  
	#endif

	// Clean up dynamically allocated arrays  
    delete[] M;  
    delete[] wi;  
    delete[] xi;  
    delete[] pi;  
}


#ifdef ENABLE_DATA_TRANSFER
void USERFUNCTION_EXPORT transferFunction(Real *result, int size, double *_CellIndex) // user field function retrieving reaction data
{
	for (unsigned int i = 0; i != size; ++i) {  
        auto it = transferMap.find(_CellIndex[i]);  
        if (it != transferMap.end())  
            result[i] = it->second;  
        else
			result[i] = 0;
    } 
}
#endif

/* Register user-code functions */
void uclib()
{
	ucfunc((void *)reactionRateModify, (char *)"IndividualReactionRate", (char *)"reacRateModify");
	
	#ifdef ENABLE_DATA_TRANSFER
	ucfunc((void *)transferFunction, (char *)"ScalarFieldFunction", (char *)"transferData");
	ucarg((void *)transferFunction, (char *)"Cell", (char *)"$LocalCellIndex", sizeof(Real));
	#endif
}

